@@ -1,3 +1,4 @@
+// -*- Mode: C++; c-basic-offset: 2; indent-tabs-mode: nil -*-
 // Copyright (c) 2005, Google Inc.
 // All rights reserved.
 //
@@ -1240,12 +1241,24 @@ REGISTER_OBJ_MAKER(nesting_i1, Nesting::Inner* p = &((new Nesting())->i1);)
 REGISTER_OBJ_MAKER(nesting_i2, Nesting::Inner* p = &((new Nesting())->i2);)
 REGISTER_OBJ_MAKER(nesting_i3, Nesting::Inner* p = &((new Nesting())->i3);)
 
+void (* volatile init_forcer)(...);
+
 // allocate many objects reachable from global data
 static void TestHeapLeakCheckerLiveness() {
   live_leak_mutable.ptr = new(initialized) char[77];
   live_leak_templ_mutable.ptr = new(initialized) Array<char>();
   live_leak_templ_mutable.val = Array<char>();
 
+  // smart compiler may see that live_leak_mutable is not used
+  // anywhere so .ptr assignment is not used.
+  //
+  // We force compiler to assume that it is used by having function
+  // variable (set to 0 which hopefully won't be known to compiler)
+  // which gets address of those objects. So compiler has to assume
+  // that .ptr is used.
+  if (init_forcer) {
+    init_forcer(&live_leak_mutable, &live_leak_templ_mutable);
+  }
   TestObjMakers();
 }
 
@@ -1266,7 +1279,7 @@ static void* Mmapper(uintptr_t* addr_after_mmap_call) {
 // On PPC64 the stacktrace returned by GetStatcTrace contains the function
 // address from .text segment while function pointers points to ODP entries.
 // The following code decodes the ODP to get the actual symbol address.
-#if defined(__linux) && defined(__PPC64__)
+#if defined(__linux) && defined(__PPC64__) && (_CALL_ELF != 2)
 static inline uintptr_t GetFunctionAddress (void* (*func)(uintptr_t*))
 {
   struct odp_entry_t {
@@ -1325,8 +1338,8 @@ static void* Mallocer(uintptr_t* addr_after_malloc_call) {
   return r;
 }
 
-// to trick complier into preventing inlining
-static void* (*mallocer_addr)(uintptr_t* addr) = &Mallocer;
+// to trick compiler into preventing inlining
+static void* (* volatile mallocer_addr)(uintptr_t* addr) = &Mallocer;
 
 // non-static for friendship with HeapProfiler
 // TODO(maxim): expand this test to include
