@@ -70,8 +70,7 @@ int ThreadCache::thread_heap_count_ = 0;
 ThreadCache* ThreadCache::next_memory_steal_ = NULL;
 #ifdef HAVE_TLS
 __thread ThreadCache::ThreadLocalData ThreadCache::threadlocal_data_
-    ATTR_INITIAL_EXEC
-    = {0, 0};
+    ATTR_INITIAL_EXEC CACHELINE_ALIGNED;
 #endif
 bool ThreadCache::tsd_inited_ = false;
 pthread_key_t ThreadCache::heap_key_;
@@ -84,7 +83,7 @@ void ThreadCache::Init(pthread_t tid) {
   if (max_size_ == 0) {
     // There isn't enough memory to go around.  Just give the minimum to
     // this thread.
-    max_size_ = kMinThreadCacheSize;
+    SetMaxSize(kMinThreadCacheSize);
 
     // Take unclaimed_cache_space_ negative.
     unclaimed_cache_space_ -= kMinThreadCacheSize;
@@ -95,8 +94,8 @@ void ThreadCache::Init(pthread_t tid) {
   prev_ = NULL;
   tid_  = tid;
   in_setspecific_ = false;
-  for (size_t cl = 0; cl < kNumClasses; ++cl) {
-    list_[cl].Init();
+  for (uint32 cl = 0; cl < Static::num_size_classes(); ++cl) {
+    list_[cl].Init(Static::sizemap()->class_to_size(cl));
   }
 
   uint32_t sampler_seed;
@@ -106,7 +105,7 @@ void ThreadCache::Init(pthread_t tid) {
 
 void ThreadCache::Cleanup() {
   // Put unused memory back into central cache
-  for (int cl = 0; cl < kNumClasses; ++cl) {
+  for (uint32 cl = 0; cl < Static::num_size_classes(); ++cl) {
     if (list_[cl].length() > 0) {
       ReleaseToCentralCache(&list_[cl], cl, list_[cl].length());
     }
@@ -115,7 +114,8 @@ void ThreadCache::Cleanup() {
 
 // Remove some objects of class "cl" from central cache and add to thread heap.
 // On success, return the first object for immediate use; otherwise return NULL.
-void* ThreadCache::FetchFromCentralCache(size_t cl, size_t byte_size) {
+void* ThreadCache::FetchFromCentralCache(uint32 cl, int32_t byte_size,
+                                         void *(*oom_handler)(size_t size)) {
   FreeList* list = &list_[cl];
   ASSERT(list->empty());
   const int batch_size = Static::sizemap()->num_objects_to_move(cl);
@@ -125,7 +125,12 @@ void* ThreadCache::FetchFromCentralCache(size_t cl, size_t byte_size) {
   int fetch_count = Static::central_cache()[cl].RemoveRange(
       &start, &end, num_to_move);
 
-  ASSERT((start == NULL) == (fetch_count == 0));
+  if (fetch_count == 0) {
+    ASSERT(start == NULL);
+    return oom_handler(byte_size);
+  }
+  ASSERT(start != NULL);
+
   if (--fetch_count >= 0) {
     size_ += byte_size * fetch_count;
     list->PushRange(fetch_count, SLL_Next(start), end);
@@ -152,7 +157,9 @@ void* ThreadCache::FetchFromCentralCache(size_t cl, size_t byte_size) {
   return start;
 }
 
-void ThreadCache::ListTooLong(FreeList* list, size_t cl) {
+void ThreadCache::ListTooLong(FreeList* list, uint32 cl) {
+  size_ += list->object_size();
+
   const int batch_size = Static::sizemap()->num_objects_to_move(cl);
   ReleaseToCentralCache(list, cl, batch_size);
 
@@ -174,10 +181,14 @@ void ThreadCache::ListTooLong(FreeList* list, size_t cl) {
       list->set_length_overages(0);
     }
   }
+
+  if (PREDICT_FALSE(size_ > max_size_)) {
+    Scavenge();
+  }
 }
 
 // Remove some objects of class "cl" from thread heap and add to central cache
-void ThreadCache::ReleaseToCentralCache(FreeList* src, size_t cl, int N) {
+void ThreadCache::ReleaseToCentralCache(FreeList* src, uint32 cl, int N) {
   ASSERT(src == &list_[cl]);
   if (N > src->length()) N = src->length();
   size_t delta_bytes = N * Static::sizemap()->ByteSizeForClass(cl);
@@ -205,8 +216,7 @@ void ThreadCache::Scavenge() {
   // that situation by dropping L/2 nodes from the free list.  This
   // may not release much memory, but if so we will call scavenge again
   // pretty soon and the low-water marks will be high on that call.
-  //int64 start = CycleClock::Now();
-  for (int cl = 0; cl < kNumClasses; cl++) {
+  for (int cl = 0; cl < Static::num_size_classes(); cl++) {
     FreeList* list = &list_[cl];
     const int lowmark = list->lowwatermark();
     if (lowmark > 0) {
@@ -241,7 +251,7 @@ void ThreadCache::IncreaseCacheLimitLocked() {
   if (unclaimed_cache_space_ > 0) {
     // Possibly make unclaimed_cache_space_ negative.
     unclaimed_cache_space_ -= kStealAmount;
-    max_size_ += kStealAmount;
+    SetMaxSize(max_size_ + kStealAmount);
     return;
   }
   // Don't hold pageheap_lock too long.  Try to steal from 10 other
@@ -259,8 +269,8 @@ void ThreadCache::IncreaseCacheLimitLocked() {
         next_memory_steal_->max_size_ <= kMinThreadCacheSize) {
       continue;
     }
-    next_memory_steal_->max_size_ -= kStealAmount;
-    max_size_ += kStealAmount;
+    next_memory_steal_->SetMaxSize(next_memory_steal_->max_size_ - kStealAmount);
+    SetMaxSize(max_size_ + kStealAmount);
 
     next_memory_steal_ = next_memory_steal_->next_;
     return;
@@ -272,8 +282,11 @@ int ThreadCache::GetSamplePeriod() {
 }
 
 void ThreadCache::InitModule() {
-  SpinLockHolder h(Static::pageheap_lock());
-  if (!phinited) {
+  {
+    SpinLockHolder h(Static::pageheap_lock());
+    if (phinited) {
+      return;
+    }
     const char *tcb = TCMallocGetenvSafe("TCMALLOC_MAX_TOTAL_THREAD_CACHE_BYTES");
     if (tcb) {
       set_overall_thread_cache_size(strtoll(tcb, NULL, 10));
@@ -282,6 +295,10 @@ void ThreadCache::InitModule() {
     threadcache_allocator.Init();
     phinited = 1;
   }
+
+  // We do "late" part of initialization without holding lock since
+  // there is chance it'll recurse into malloc
+  Static::InitLateMaybeRecursive();
 }
 
 void ThreadCache::InitTSD() {
@@ -303,8 +320,35 @@ void ThreadCache::InitTSD() {
 }
 
 ThreadCache* ThreadCache::CreateCacheIfNecessary() {
+  if (!tsd_inited_) {
+#ifndef NDEBUG
+    // tests that freeing nullptr very early is working
+    free(NULL);
+#endif
+
+    InitModule();
+  }
+
   // Initialize per-thread data if necessary
   ThreadCache* heap = NULL;
+
+  bool seach_condition = true;
+#ifdef HAVE_TLS
+  static __thread ThreadCache** current_heap_ptr;
+  if (tsd_inited_) {
+    // In most common case we're avoiding expensive linear search
+    // through all heaps (see below). Working TLS enables faster
+    // protection from malloc recursion in pthread_setspecific
+    seach_condition = false;
+
+    if (current_heap_ptr != NULL) {
+      // we're being recursively called by pthread_setspecific below.
+      return *current_heap_ptr;
+    }
+    current_heap_ptr = &heap;
+  }
+#endif
+
   {
     SpinLockHolder h(Static::pageheap_lock());
     // On some old glibc's, and on freebsd's libc (as of freebsd 8.1),
@@ -328,10 +372,12 @@ ThreadCache* ThreadCache::CreateCacheIfNecessary() {
     // This may be a recursive malloc call from pthread_setspecific()
     // In that case, the heap for this thread has already been created
     // and added to the linked list.  So we search for that first.
-    for (ThreadCache* h = thread_heaps_; h != NULL; h = h->next_) {
-      if (h->tid_ == me) {
-        heap = h;
-        break;
+    if (seach_condition) {
+      for (ThreadCache* h = thread_heaps_; h != NULL; h = h->next_) {
+        if (h->tid_ == me) {
+          heap = h;
+          break;
+        }
       }
     }
 
@@ -348,10 +394,13 @@ ThreadCache* ThreadCache::CreateCacheIfNecessary() {
 #ifdef HAVE_TLS
     // Also keep a copy in __thread for faster retrieval
     threadlocal_data_.heap = heap;
-    SetMinSizeForSlowPath(kMaxSize + 1);
+    threadlocal_data_.fast_path_heap = heap;
 #endif
     heap->in_setspecific_ = false;
   }
+#ifdef HAVE_TLS
+  current_heap_ptr = NULL;
+#endif
   return heap;
 }
 
@@ -384,7 +433,7 @@ void ThreadCache::BecomeIdle() {
 #ifdef HAVE_TLS
   // Also update the copy in __thread
   threadlocal_data_.heap = NULL;
-  SetMinSizeForSlowPath(0);
+  threadlocal_data_.fast_path_heap = NULL;
 #endif
   heap->in_setspecific_ = false;
   if (GetThreadHeap() == heap) {
@@ -397,6 +446,12 @@ void ThreadCache::BecomeIdle() {
   DeleteCache(heap);
 }
 
+void ThreadCache::BecomeTemporarilyIdle() {
+  ThreadCache* heap = GetCacheIfPresent();
+  if (heap)
+    heap->Cleanup();
+}
+
 void ThreadCache::DestroyThreadCache(void* ptr) {
   // Note that "ptr" cannot be NULL since pthread promises not
   // to invoke the destructor on NULL values, but for safety,
@@ -405,7 +460,7 @@ void ThreadCache::DestroyThreadCache(void* ptr) {
 #ifdef HAVE_TLS
   // Prevent fast path of GetThreadHeap() from returning heap.
   threadlocal_data_.heap = NULL;
-  SetMinSizeForSlowPath(0);
+  threadlocal_data_.fast_path_heap = NULL;
 #endif
   DeleteCache(reinterpret_cast<ThreadCache*>(ptr));
 }
@@ -443,7 +498,7 @@ void ThreadCache::RecomputePerThreadCacheSize() {
     // Increasing the total cache size should not circumvent the
     // slow-start growth of max_size_.
     if (ratio < 1.0) {
-        h->max_size_ = static_cast<size_t>(h->max_size_ * ratio);
+      h->SetMaxSize(h->max_size_ * ratio);
     }
     claimed += h->max_size_;
   }
@@ -455,7 +510,7 @@ void ThreadCache::GetThreadStats(uint64_t* total_bytes, uint64_t* class_count) {
   for (ThreadCache* h = thread_heaps_; h != NULL; h = h->next_) {
     *total_bytes += h->Size();
     if (class_count) {
-      for (int cl = 0; cl < kNumClasses; ++cl) {
+      for (int cl = 0; cl < Static::num_size_classes(); ++cl) {
         class_count[cl] += h->freelist_length(cl);
       }
     }
